<?php
/**
 * @file
 * Drupal Inspect module stack trace inspection class
 */

/**
 * Stack trace inspector
 * @static
 * @singleton
 * @category Drupal
 * @package Developer
 * @subpackage Inspect
 */
class InspectTrace extends Inspect {
  /**
   * Prevents erratic instantiation of this entirely static class
   */
  private function  __construct() {}

  /**
   * Default sub var recursion depth.
   *
   * Overrides Inspect::DEPTH_DEFAULT - when methods of this class uses methods of Inspect (using static::DEPTH_DEFAULT).
   *
   * @type integer
   */
  const DEPTH_DEFAULT = 10;

  /**
   * Minimum string truncation.
   *
   * Overrides Inspect::TRUNCATE_MIN - when methods of this class uses methods of Inspect (using static::TRUNCATE_MIN).
   *
   * @type integer
   */
  const TRUNCATE_MIN = 100000;

  /**
   * Overrides Inspect::TRUNCATE_DEFAULT - when methods of this class uses methods of Inspect (using static::TRUNCATE_DEFAULT).
   *
   * @type integer
   */
  const TRUNCATE_DEFAULT = 1000;

  /**
   * Absolute maximum stack frame depth
   *
   * @type integer
   */
  const TRACE_LIMIT_MAX = 100;

  /**
   * Default trace options.
   *
   * Adds overrides options of Inspect::$_options[get/log|syslog|file].
   *
   *   Options:
   *   - (string) category: inspect trace
   *   - (string) trace_spacer: spacer between frames
   *   - (string) pre_indent: indentation of object/array buckets
   *   - (array) delimiters: see Inspect
   *
   * @var array $_optionsTrace
   */
  protected static $_optionsTrace = array(
    'get' => array(
      'category' => 'inspect trace',
      'trace_spacer' => '- - - - - - - - - - - - - - - - - - - - - - - - -',
      'pre_indent' => '  ',
      'delimiters' => array(
        "\n", // Before first bucket.
        "\n", // Before later bucket.
        "\n  " // Before last bucket.
      ),
    ),
    'log' => array(
      'category' => 'inspect trace',
      'trace_spacer' => '- - - - - - - - - - - - - - - - - - - - - - - - -',
      'pre_indent' => '  ',
      'delimiters' => array(
        "\n", // Before first bucket.
        "\n", // Before later bucket.
        "\n  " // Before last bucket.
      ),
    ),
    'syslog' => array(
      'category' => 'inspect trace',
      'trace_spacer' => '- - -',
      'pre_indent' => ' ',
      'delimiters' => array(
        '\\n', // Before first bucket.
        '\\n', // Before later bucket.
        '\\n ' // Before last bucket.
      ),
    ),
    'file' => array(
      'category' => 'inspect trace',
      'trace_spacer' => '- - - - - - - - - - - - - - - - - - - - - - - - -',
      'pre_indent' => '  ',
      'delimiters' => array(
        "\n", // Before first bucket.
        "\n", // Before later bucket.
        "\n  " // Before last bucket.
      ),
    ),
  );

  /**
   * @see InspectTrace::_init()
   * @var boolean|NULL $_init
   */
  protected static $_init;

  /**
   * Merges own options with parent options.
   *
   * @see Inspect::_options()
   * @return boolean
   */
  protected static function _init() {
    if (!parent::$_init) {
      parent::_init();
    }
    $o =& self::$_options; // Inspect's options.

    $oTrc =& self::$_optionsTrace; // Own options, to be merged with parent's.

    //  If syslog: use syslog options as log options.
    if (self::$_syslog) {
      $oTrc['log'] = $oTrc['syslog'];
    }
    //  If drush/CLI: use default file options instead of get options.
    if (drupal_is_cli()) {
      $oTrc['get'] = $oTrc['file'];
    }
    //  Merge all types of options.
    $oTrc['get'] = array_merge($o['get'], $oTrc['get']);
    $oTrc['log'] = array_merge($o['log'], $oTrc['log']);
    $oTrc['file'] = array_merge($o['file'], $oTrc['file']);

    return (self::$_init = TRUE);
  }

  /**
   * Analyzes exception and creates derivate as formatted string.
   *
   * @param string $target
   *  - allowed values: get|log|file
   * @param Exception|falsy $xc
   *  - if falsy: create new trace
   * @param array &$opts
   *  - resolved options
   * @return string|boolean
   *  - FALSE on error (other error than exceeding max. output length)
   */
  protected static function _trc($target, $xc, &$opts) {
    //  Secure limit option.
    if (!isset($opts['limit']) || $opts['limit'] > static::TRACE_LIMIT_MAX) {
      $opts['limit'] = variable_get('inspect_trclimitdef', 5);
    }
    elseif ($opts['limit'] < 1) {
      $opts['limit'] = 1;
    }

    try {
      //  Received Exception, by arg.
      if ($xc) {
        if (is_object($xc) && (($u = get_class($xc)) === 'Exception' || is_subclass_of($xc, 'Exception'))) {
          $xcClass = $u;
          $trace = $xc->getTrace();
          if (count($trace) > $opts['limit']) {
            array_splice($trace, $opts['limit']);
          }
        }
        else {
          throw new Exception(
            'arg $exception type[' . (!is_object($xc) ? gettype($xc) : get_class($xc)) . '] is not Exception or falsy',
            Inspect::ERROR_USER
          );
        }
      }
      //  Create trace, if none given by arg.
      else {
        $xcClass = '';
        $trace = debug_backtrace();
        //  Remove top levels on synthetic trace, first of all this method.
        array_shift($trace);
        //  If called via function instead of class method: remove class method call.
        if (isset($trace[0]['class']) && strtolower($trace[0]['class']) === 'inspecttrace' // Deliberately not drupal_strtolower().
          && isset($trace[1]['function']) && strtolower(substr($trace[1]['function'], 0, 13)) === 'inspect_trace' // Deliberately not drupal_...().
        ) {
          array_shift($trace);
        }
        //  Enforce the trace limit.
        if (count($trace) > $opts['limit']) {
          array_splice($trace, $opts['limit'] + 1); // plus one because we need the bucket holding the initial event
        }
      }

      //  Secure output max.
      self::$_outputLength = 0;
      self::$_outputMax = $target == 'log' ? (self::$_outputMax_log - static::LOG_MAX_MARGIN) : self::$_outputMax_get; // Yes, file uses get max.

      //  Format as string.
      $hidePaths = $opts['hide_paths'];
      $delim = $opts['delimiters'][1];
      $preIndent = $opts['pre_indent'];
      $frameSpacer = $opts['trace_spacer'];
      $sFunc = '';
      //  If exception: resolve its origin.
      if ($xcClass) {
        $sTrc = 'Exception (' . $xcClass . ') - code: ' . intval($xc->getCode()) . $delim;
        $file = trim($xc->getFile());
        $line = (($u = $xc->getLine()) ? trim($u) : '?');
        $message = 'message: ' . ($u = $xc->getMessage()) . $delim;
        //  If severe: pass exception message to overall message, if that is empty.
        if ($opts['severity'] != WATCHDOG_DEBUG && empty($opts['message'])) {
          $opts['message'] = $u;
          if (drupal_strlen($u = htmlspecialchars(
            $opts['message'],
            ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
            'UTF-8',
            FALSE // No double encoding
          )) > 255) {
            $u = drupal_substr($u, 0, 255);
          }
          $opts['message'] = $u;
        }
      }
      else {
        $sTrc = 'Backtrace' . $delim;
        //  Find inspect method/function, which commanded creation of synthetic trace.
        $frame = array_shift($trace);
        if (isset($frame['class'])) {
          $sFunc = 'static method: ' . $frame['class'] . '::' . (isset($frame['function']) ? ($frame['function'] . '()') : '') . $delim;
        }
        elseif (isset($frame['function'])) {
          $sFunc = 'function: ' . $frame['function'] . '()' . $delim;
        }
        if (isset($frame['file'])) {
          $file = trim($frame['file']);
        }
        $line = (isset($frame['line']) ? trim($frame['line']) : '?');
        $message = '';
      }
      $file = ($le = drupal_strlen($file)) ? (
        $hidePaths ? basename($file) : (
          (drupal_strlen($noRoot = preg_replace(self::$_paths, '', $file)) < $le ? ('document_root/' . $noRoot) : $file)
        )
      ) : 'unknown';
      $sTrc .= 'file: ' . str_replace('\\', '/', $file) . $delim
        . 'line: ' . $line . $delim
        . $sFunc
        . $message;

      $sTrcEnd = 'END ' . $frameSpacer;

      //  Iterate stack frames.
      $sTrcLength = strlen($sTrc) + strlen($sTrcEnd); // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
      $nFrame = -1;
      foreach ($trace as &$frame) {
        $sFrm = (++$nFrame) . ' ' . $frameSpacer . $delim;
        //  File and line.
        $file = isset($frame['file']) && ($le = drupal_strlen($u = trim($frame['file']))) ? (
          $hidePaths ? basename($u) : (
            (drupal_strlen($noRoot = preg_replace(self::$_paths, '', $u)) < $le ? ('document_root/' . $noRoot) : $u)
          )
        ) : 'unknown';
        $sFrm .= 'file: ' . str_replace('\\', '/', $file) . $delim . 'line: ' . (isset($frame['line']) ? trim($frame['line']) : '?') . $delim;
        //  Class, object, function, type.
        $sFunc = isset($frame['function']) && strlen($u = $frame['function']) ? $u : ''; // Deliberately not drupal_strlen().
        $sCls = isset($frame['class']) && strlen($u = $frame['class']) ? $u : ''; // Deliberately not drupal_strlen().
        $sType = isset($frame['type']) ? trim($frame['type']) : '';
        $sObj = '';
        if (isset($frame['object'])) {
          if (($o = $frame['object'])) {
            //  We dont know if class bucket is present when object is, but using class is cheaper.
            $sObj = $sCls ? $sCls : get_class($o);
            if ($sFunc) {
              $sFunc = 'method: (' . $sObj . ')' . ($sType ? $sType : '->') . $sFunc;
            }
            else { // Is this possible at all? Simply object, no method call??
              $sObj = 'object (' . $sObj . ')';
            }
          }
        }
        elseif ($sFunc) {
          $sFunc = !$sCls ? ('function: ' . $sFunc) : ('static method: ' . $sCls . ($sType ? $sType : '::') . $sFunc);
          //  $frame['object'] doesnt exist.
        }
        //  else {
        //    $frame['object'] doesnt exist
        //    $frame['function'] may exist, containing non-empty string
        //  }
        $sFrm .= ($sFunc ? $sFunc : $sObj) . $delim;
        //  Args.
        if (isset($frame['args'])) {
          $le = count($args = $frame['args']);
          $sArgs = '';
          for ($i = 0; $i < $le; $i++) {
            $sArgs .= $delim . $preIndent
              . self::_inspect(
                $args[$i],
                $opts,
                0 // recursion counter
              );
          }
          $sFrm .= 'args (' . $le . ')'
            . (!$le ? '' : (': ' . $sArgs) )
            . $delim;
        }
        //  Skip current frame if total output now exceeds max.
        if (($sTrcLength += strlen($sFrm)) > self::$_outputMax) { // Deliberately not drupal_strlen(); need 'physical' length, not (possibly shorter) multibyte length.
          $opts['warning_limit_reduced'] = $nFrame;
          break;
        }
        else {
          $sTrc .= $sFrm;
        }
      }
      return $sTrc . $sTrcEnd;
    }
    catch (Exception $xc1) {
      //  Inspect::_inspect() may abort too.
      if (($errorCode = $xc1->getCode()) == Inspect::ERROR_OUTPUTLENGTH) {
        $opts['warning_limit_reduced'] = $nFrame;
        return $sTrc . $sTrcEnd;
      }
      //  Other error.
      if ($opts['severity'] >= WATCHDOG_DEBUG) {
        $opts['severity'] = WATCHDOG_WARNING;
      }
      return htmlspecialchars(
        str_replace($opts['needles'], $opts['replacers'], $xc1->getMessage()) . (!$errorCode ? '' : (' (error ' . $errorCode . ')')),
        ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
        'UTF-8',
        FALSE // No double encoding
      );
    }
  }

  /**
   * Inspect and log stack trace to database (or syslog), if permitted.
   *
   * Logging an error message and/or exception message is always allowed, if severity is more severe than WATCHDOG_DEBUG.
   *
   *   Integer value of $options (limit):
   *   - maximum stack frame depth
   *   - default: 5 (configurable in the settings page)
   *   - max: 100 (InspectTrace::TRACE_LIMIT_MAX)
   *
   *   Array/object value of $options are like Inspect::log()'s options, except:
   *   - (integer) limit (default 5, max 100)
   *   - (string) trace_spacer: spacer between frames (default hyphen dotted line, length 49)
   *   - (string) category: inspect trace
   *   - (boolean) file_line: is not used by tracer
   *
   * See inspect() for specifics of variable analysis.
   *
   * Default watchdog severity: WATCHDOG_DEBUG.
   *
   * Executes variable analysis within try-catch.
   *
   * @see Inspect::log()
   * @param Exception|falsy $exception
   *   - default: NULL (~ create new backtrace)
   *   - if Exception, trace that
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to limit (maximum stack frame depth)
   *   - string: interprets to message
   *   - other type: ignored
   * @return boolean|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   */
  public static function log($exception = NULL, $options = NULL) {
    //  Logging a trace is always permitted, when severity is less than WATCHDOG_DEBUG (more severe);
    //  however only the exception message and/or $options message will be logged.
    //  Thus we have to resolve options before checking permission; to establish severity.
    if (!self::$_init) {
      self::_init();
    }
    $opts = self::_options('log', self::$_optionsTrace['log'], $options, 'limit');
    if (!self::_permit('log')) {
      if (($severity = $opts['severity']) < WATCHDOG_DEBUG) {
        if ($exception && is_object($exception)) {
          $em = htmlspecialchars(
            str_replace($opts['needles'], $opts['replacers'], $exception->getMessage()),
            ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
            'UTF-8',
            FALSE // No double encoding
          ) . $opts['newline'] . '@' . basename($exception->getFile()) . ':' . $exception->getLine();
        }
        else {
          $em = self::_fileLine(NULL, TRUE, empty($opts['wrappers']) ? 0 : $opts['wrappers']);
        }
        $ms = !empty($opts['message']) ? htmlspecialchars(
          str_replace($opts['needles'], $opts['replacers'], $opts['message']),
          ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
          'UTF-8',
          FALSE // No double encoding
        ) : '';
        return self::_log(
          $ms . ($ms && $em ? (':' . $opts['newline']) : '') . $em,
          $opts['category'],
          $opts['severity']
        );
      }
      return NULL;
    }

    if (!($trace = self::_trc('log', $exception, $opts))) {
      return FALSE;
    }
    $nl = $opts['newline'];
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag'])) {
      $tagStart = '<' . $u . ' class="module-inspect-trace">';
      $tagEnd = '</' . $u . '>';
    }
    return self::_log(
      $tagStart
        . ( empty($opts['message']) ?
          ('[Inspect trace - ' . join(':', self::$_sessionCounters) . ':' . (++self::$_logNo)
            . ' - limit:' . $opts['limit'] . '|depth:' . $opts['depth'] . ']|truncate:' . $opts['truncate'] . ']') :
          ('[' . join(':', self::$_sessionCounters) . ':' . (++self::$_logNo) . '] '
            . $opts['message'] . ':' . $nl . '[Inspect trace - limit:'
            . $opts['limit'] . (empty($opts['warning_limit_reduced']) ? '' : '(reduced to ' . $opts['warning_limit_reduced'] . ')')
            . '|depth:' . $opts['depth'] . ']|truncate:' . $opts['truncate'] . ']')
        )
        . $nl
        . $trace
        . $tagEnd,
      $opts['category'],
      $opts['severity']
    );
  }

  /**
   * Inspect and log stack trace to file, if permitted.
   *
   * Filing an error message and/or exception message is always allowed, if severity is more severe than WATCHDOG_DEBUG.
   *
   *   Array/object value of $options are like InspectTrace::log()'s options, except:
   *   - (boolean) by_user: log to user specific file instead of common file
   *
   * @see Inspect::file()
   * @param Exception|falsy $exception
   *   - default: NULL (~ create new backtrace)
   *   - if Exception, trace that
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to limit (maximum stack frame depth)
   *   - string: interprets to message
   *   - other type: ignored
   * @return boolean|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error (other error than exceeding max. output length)
   */
  public static function file($exception = NULL, $options = NULL) {
    //  Filing a trace is always permitted, when severity is less than WATCHDOG_DEBUG (more severe);
    //  however only the exception message and/or $options message will be logged.
    //  Thus we have to resolve options before checking permission; to establish severity.
    if (!self::$_init) {
      self::_init();
    }
    $opts = self::_options('file', self::$_optionsTrace['file'], $options, 'limit');
    if (!self::_permit('file')) {
      if (($severity = $opts['severity']) < WATCHDOG_DEBUG) {
        if ($exception && is_object($exception)) {
          $em = htmlspecialchars(
            str_replace(
              $opts['needles'],
              $opts['replacers'],
              $exception->getMessage()
            ),
            ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
            'UTF-8',
            FALSE // No double encoding
          ) . $opts['newline'] . '@' . basename($exception->getFile()) . ':' . $exception->getLine();
        }
        else {
          $em = self::_fileLine(NULL, TRUE, empty($opts['wrappers']) ? 0 : $opts['wrappers']);
        }
        $ms = !empty($opts['message']) ? htmlspecialchars(
          str_replace(
            $opts['needles'],
            $opts['replacers'],
            $opts['message']
          ),
          ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
          'UTF-8',
          FALSE // No double encoding
        ) : '';
        return self::_file(
          $ms . ($ms && $em ? (':' . $opts['newline']) : '') . $em,
          $opts['category'],
          $opts['severity'],
          empty($opts['by_user']) ? FALSE : $opts['by_user']
        );
      }
      return NULL;
    }

    if (!($trace = self::_trc('file', $exception, $opts))) {
      return FALSE;
    }

    $nl = $opts['newline'];
    return self::_file(
      ( empty($opts['message']) ? '' : ($opts['message'] . ':' . $nl) )
        . '[Inspect trace - ' . join(':', self::$_sessionCounters) . ':' . (++self::$_logNo)
          . ' - limit:' . $opts['limit'] . (empty($opts['warning_limit_reduced']) ? '' : '(reduced to ' . $opts['warning_limit_reduced'] . ')')
          . '|depth:' . $opts['depth'] . ']|truncate:' . $opts['truncate'] . ']'
          . $nl
        . $trace,
      $opts['category'],
      $opts['severity'],
      empty($opts['by_user']) ? FALSE : $opts['by_user']
    );
  }

  /**
   * Inspect stack trace and get the output as string, if permitted.
   *
   * @see InspectTrace::log()
   * @param Exception|falsy $exception
   *   - default: NULL (~ create new backtrace)
   *   - if Exception, trace that
   * @param mixed $options
   *   - default: NULL
   *   - array|object: list of options
   *   - integer: interprets to limit (maximum stack frame depth)
   *   - string: interprets to message
   *   - other type: ignored
   * @return string
   *   - empty if user isnt permitted to get inspections, or tracing failed (other error than exceeding max. output length)
   */
  public static function get($exception = NULL, $options = NULL) {
    if (!self::_permit('get')) {
      return '';
    }
    if (!self::$_init) {
      self::_init();
    }
    $opts = self::_options('get', self::$_optionsTrace['get'], $options, 'limit');

    if (!($trace = self::_trc('get', $exception, $opts))) {
      return '';
    }

    $nl = $opts['newline'];
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag'])) {
      $tagStart = '<' . $u . ' class="module-inspect-trace">';
      $tagEnd = '</' . $u . '>';
    }
    return $tagStart
      . ( empty($opts['message']) ?
        ('[Inspect trace - ' . join(':', self::$_sessionCounters) . ':' . (++self::$_logNo)
          . ' - limit:' . $opts['limit'] . '|depth:' . $opts['depth'] . ']|truncate:' . $opts['truncate'] . ']') :
        ('[' . join(':', self::$_sessionCounters) . ':' . (++self::$_logNo) . '] '
          . $opts['message'] . ':' . $nl . '[Inspect trace - limit:'
          . $opts['limit'] . (empty($opts['warning_limit_reduced']) ? '' : '(reduced to ' . $opts['warning_limit_reduced'] . ')')
          . '|depth:' . $opts['depth'] . ']|truncate:' . $opts['truncate'] . ']')
      )
      . $nl
      . $trace
      . $tagEnd;
  }
}